/***************************************/
/*  Use MADS http://mads.atari8.info/  */
/*  Mode: GED- (bitmap mode)           */
/***************************************/

; Drpeter's commented, centring version of no_name from AtariAge RastaConverter forum.
; After some discussion between us, additional modifications here make binary files with
; the same segment layout and kernel instructions as standard RastaConverter XEXs which
; work with older RastaSlide versions.

	icl "no_name.h"



byt2	= $14     ; zero page dummy variable for timing operations (real time clock counter)

* ---	BASIC switch OFF
	org $2000\ mva #$ff portb\ rts\ ini $2000

* ---	MAIN PROGRAM
	org $2010                                ; means that 12K boundary falls on 102 x 40 lines so no screen data offset needed here to jump the 12K ($3000) boundary
	PAD_LINES = (240-PIC_HEIGHT)/2
	PAD_BYTES = PAD_LINES*40
	IFT (PIC_HEIGHT+PAD_LINES)*40 > 8160
scr :PAD_BYTES .byte 0
	ins "output.png.mic", 0, (8160-PAD_BYTES) ; screen data going beyond 16K boundary, ($2010 + 40*204=8160 ($3FF0) so align next screen line forward by 16 bytes to $4000
	:16 .byte 0
	ins "output.png.mic" , +(8160-PAD_BYTES)
	ELS
scr :PAD_BYTES .byte 0
	ins "output.png.mic"
	EIF


	.ALIGN $0400        ; 1K boundary for display list
ant	ANTIC_PROGRAM scr,ant

fnt

	ift USESPRITES
	.ALIGN $0800        ; 2K boundary for PM data block
	.ds $0300           ; then offset $300 to start of PM data
pmg	.ds PAD_LINES		; shift PM data down to account for padding
	SPRITES             ; PM data loaded here for $500 bytes , included via SPRITES macro from 'output.png.pmg'
	;                   org *-PAD_LINES	    ; realign with 2K boundary at upper limit of PM data
	eif

main					;############ program starts here #############
* ---	init PMG

	ift USESPRITES
	mva >pmg pmbase		;missiles and players data address
	mva #$03 gractl		;enable players and missiles
	eif

	lda:cmp:req $14		;wait 1 frame

	sei			     ;stop interrupts
	mva #$00 NMIEN		;stop non-maskable interrupts too
	mva #$fe PORTB		;switch off ROM to get 16k more ram (not actually needed)


////////////////////
// RASTER PROGRAM //
////////////////////

	jmp raster_program_end  ;initialise graphics registers etc. ready for 1st frame

LOOP	lda vcount		;synchronization for the first screen line
	cmp #$02			;wait for end of 5th scan line (VCOUNT will go 0;0;1;1;2;2 etc.)
	bne LOOP

	mva #%00111110 dmactl	;ANTIC playfield DMA on/single line players/player DMA on/missile DMA on/ normal screen width
	mva <ant dlptr          ;reset dlist pointer (dlist doesn't trigger until scan line 8) necessary because JVB can't be used on 240 line display (see Altirra Hardware Manual 4.6)
	mva >ant dlptr+1		;total of 18 cycles. no PF DMA or DL DMA or PMG DMA 1st 8 scan-lines so 9 DMA cycles/line for refresh, leaving 105 cycles/line

  icl "output.png.rp.ini"   ;run rasta frame initialisation, includes 78 cycles then a final WSYNC x2 to set up horizontal screen synchronisation -> cycle 105 of 7th scanline

;--- wait 18 cycles         ;from cycle 104 to end of what would usually be DLIST/PMG DMA block @ cycle 7
	jsr _rts   ;12
	inc byt3   ;6

;--- set global offset (23 cycles)
	jsr _rts			;12
	cmp byt3\ pha:pla   ;4/3/4

;--- empty line (59 cycles)
	jsr wait54cycle
	inc byt2            ;5   now total of 7+23+59 = cycle 89, + 9 refresh cycles = cycle 98, rasta program starts will the cycle delay before line 0, taking to end of cycle 109 of 8th scan-line
						;    i.e. synchronised to start rasta kernel proper with mode E line 0, 57 cycles per line


;--- mark kernel time during padded lines	
	:PAD_LINES wait57cycle
	
  icl "output.png.rp"       ;start rasta kernel from here

raster_program_end


c0	lda #$00               ;set all playfield/background colours to black
	sta colbak			   ;avoids accidental display of data in memory above end of rasta picture
c1	lda #$00
	sta color0
c2	lda #$00
	sta color1
c3	lda #$00
	sta color2
c4	lda #$00
	sta color3
s0	lda #$03
	sta sizep0				;set all player & missile sizes to wide
	sta sizep1
	sta sizep2
	sta sizep3
	mva #$ff sizem
	sta grafm				;load missile data register with $FF (this is never refreshed as missile data receipt from ANTIC is going to be disabled on the GTIA)
	mva #$20 hposm0			;position missiles to mask borders
	mva #$28 hposm1
	mva #$d0 hposm2
	mva #$d8 hposm3
	mva #$02 gractl			;disable missile data receipt by GTIA, so $FF remains permanently loaded in GRAFM
	lda #$14				;%00010100
	sta prior				;use missiles as 5th player with priority PF0-PF1-PF2-PF3 P0-P1-P2-P3 BAK, so they (having PF0 priority) mask other players and background in borders


//--------------------
//	EXIT?					; terminate program with trigger 0, trigger 1, any key or [START] & return to BASIC
//--------------------

	lda trig0		; FIRE #0
	beq stop

	lda trig1		; FIRE #1
	beq stop

	lda consol		; START
	and #1
	beq stop

	lda skctl		; ANY KEY
	and #$04
	bne skp

stop	mva #$00 gractl		;PMG disabled
	tax
	sta:rne hposp0,x+		; zero all GTIA registers

	mva #$fd portb		;ROM & BASIC switch on
	mva #$40 nmien		;only NMI interrupts, DLI disabled
	cli			        ;IRQ enabled

	rts		         	;return to ... BASIC
skp

//--------------------

	jmp LOOP			;start new frame

;---
.MACRO wait57cycle
	:+04 nop
	cmp byt2
	:+23 nop
.ENDM
wait54cycle   
	:2 inc byt2 ;5 x 2
wait44cycle
	inc byt3  ;6
	nop       ;2
wait36cycle
	inc byt3  ;8
	jsr _rts  ;12
wait18cycle
	inc byt3  ;6
_rts	rts   ;12

byt3	brk


;---

.MACRO	ANTIC_PROGRAM
	:+8 dta $4e,a(:1+$0000+#*40)
	:+8 dta $4e,a(:1+$0140+#*40)
	:+8 dta $4e,a(:1+$0280+#*40)
	:+8 dta $4e,a(:1+$03C0+#*40)
	:+8 dta $4e,a(:1+$0500+#*40)
	:+8 dta $4e,a(:1+$0640+#*40)
	:+8 dta $4e,a(:1+$0780+#*40)
	:+8 dta $4e,a(:1+$08C0+#*40)
	:+8 dta $4e,a(:1+$0A00+#*40)
	:+8 dta $4e,a(:1+$0B40+#*40)
	:+8 dta $4e,a(:1+$0C80+#*40)
	:+8 dta $4e,a(:1+$0DC0+#*40)
	:+8 dta $4e,a(:1+$0F00+#*40)
	:+8 dta $4e,a(:1+$1040+#*40)
	:+8 dta $4e,a(:1+$1180+#*40)
	:+8 dta $4e,a(:1+$12C0+#*40)
	:+8 dta $4e,a(:1+$1400+#*40)
	:+8 dta $4e,a(:1+$1540+#*40)
	:+8 dta $4e,a(:1+$1680+#*40)
	:+8 dta $4e,a(:1+$17C0+#*40)
	:+8 dta $4e,a(:1+$1900+#*40)
	:+8 dta $4e,a(:1+$1A40+#*40)
	:+8 dta $4e,a(:1+$1B80+#*40)
	:+8 dta $4e,a(:1+$1CC0+#*40)
	:+8 dta $4e,a(:1+$1E00+#*40)
	:+4 dta $4e,a(:1+$1F40+#*40) ;approaching 16K boundary, so only 4 lines
	:+4 dta $4e,a(:1+$1FF0+#*40) ;moved across 16K boundary, so jumped forward by 16 bytes
	:+8 dta $4e,a(:1+$2090+#*40)
	:+8 dta $4e,a(:1+$21D0+#*40)
	:+8 dta $4e,a(:1+$2310+#*40)
	:+8 dta $4e,a(:1+$2450+#*40) ;total of 240 lines of ANTIC mode E with LMS on each line
	dta $41,a(:2)                ;JVB jump to start (see Altirra Hardware Manual 4.6 for need to handle this with care for 240 line display)
.ENDM

CL

;---
	run main
;---

	opt l-

.MACRO	SPRITES
	icl "output.png.pmg"
.ENDM

USESPRITES = 1

