//
// RastaSlide 2.0 - Chris Hutt
//

// This started life as a quick and dirty script, and it will become
// obvious I haven't learned much GO, so I can't recommend much here
// as any great example of how to do things.
// Freeware. No warranty.
// 32-bit binary made with the options below to reduce some bloat
// set GOARCH=386
// go build -trimpath -ldflags="-s -w" rslide.go

package main

import (
	"bytes"
	"compress/flate"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"os/exec"
	"path/filepath"
	"runtime"
	"strconv"
	"strings"
	"time"
	"unicode"
	// need to install git, then do "go get golang.org/x/text/runes" etc.
	// which wasn't at all obvious to a beginner
	"golang.org/x/text/runes"
	"golang.org/x/text/transform"
	"golang.org/x/text/unicode/norm"
)

const ignoreErr = true
const checkErr = false

var err error

// global - don't want to keep passing these between lots of function calls
// nobody seems to like globals much any more... but I'm not sure how to rewrite it
var fld, pos, reg, addr int
var regval []uint64 = make([]uint64, 4) // no register to map or A,X,Y
var regold []uint64 = make([]uint64, 4)
var addrval []uint64 = make([]uint64, 16) // only need enough for player and colour registers
var addrold []uint64 = make([]uint64, 16) // and a couple of others
var ld, st, wt, hpos bool
var colmask uint64

//
// main
//

func main() {

	var path, fails string
	var arg int
	var zini []byte
	var zopts []string

	fmt.Println("RastaSlide - RastaConverter Slide Show Maker - Version 2.0")

	// exit if no command and arguments
	if len(os.Args) < 3 {
		fmt.Println("rslide usage:")
		fmt.Println("rslide add 128/256 files/folders... - Add RastaConverter pictures to slides")
		fmt.Println("rslide bld outfile                  - Build slides into RastaSlide show program and disks")
		fmt.Println("rslide dis [opt] infile             - Disassemble RastaConverter picture to output files")
		fmt.Println("rslide ext pic/sld infile.atr/xex   - Extract all pictures or slides from RastaSlide show")
		fmt.Println("rslide opt 128/256 infile.rp        - Optimize RastaConverter kernel and .ini to .opt files")
		fmt.Println()
		fmt.Println("See Generator folder for output and kernel files")
		fmt.Println("128/256 - Optimize to 128 colours or leave at 256(only all visible with a Sophia chip)")
		return
	}

	// path where this program is will be used for working files produced by it
	path, err = filepath.Abs(filepath.Dir(os.Args[0]))
	check(err)
	// change directory works for network folders too in GO, unlike Windows command line
	os.Chdir(path)
	check(err)

	// rslide command argument
	switch os.Args[1] {
	case "add":
		// make viewer header for picture assembly. can be removed once development over
		winCmd(checkErr, "mads", "viewer.asm", "-hm:viewer.h", "-o:output.dat", "-d:PIC_QTY=0")

		// get zopfli option settings. NB even with the faster and parallel KrzYmod zopfli
		// fork used, aggressive settings will take a very long time
		zini, err = ioutil.ReadFile("zopfli.ini")
		check(err)
		// deflate method and use threads where possible (these small files only split into 2 or 3 threads)
		zopts = append(strings.Split(string(zini), " "), "--t"+strconv.Itoa(runtime.NumCPU()), "--deflate")
		fmt.Println(zopts)

		// add all passed files
		for arg = 3; arg < len(os.Args); arg++ {

			file, _ := os.Open(os.Args[arg])
			defer file.Close()
			dir, _ := file.Readdirnames(-1)

			// file could be a directory or a normal file
			if len(dir) != 0 {
				for _, f := range dir {
					// don't even attempt other file types in the folder
					if strings.ToLower(filepath.Ext(f)) != ".xex" {
						continue
					}
					if !addSlide(filepath.Join(os.Args[arg], f), os.Args[2], zopts) {
						fails += "Failed and skipped " + filepath.Join(os.Args[arg], f) + "\n"
					}
					if fails != "" {
						fmt.Print(fails)
					}
				}
			} else {
				if !addSlide(os.Args[arg], os.Args[2], zopts) {
					fails += "Failed and skipped " + os.Args[arg] + "\n"
				}
				if fails != "" {
					fmt.Print(fails)
				}
			}
		}
		if fails == "" {
			time.Sleep(2 * time.Second)
		} else {
			os.Exit(2)
		}
	case "bld":
		buildSlides(os.Args[2])
		time.Sleep(2 * time.Second)
	case "dis":
		fmt.Println("Disassembling...")
		if os.Args[2] != "opt" {
			disAsmRasta(os.Args[2], "output")
		} else {
			disAsmRasta(os.Args[3], "output")
			fmt.Println("Optimizing...")
			optKernel("output.png.rp", "output.png.opt", "256", false)
			// make an optimized kernel with nice timing comments as the above isn't prettified
			winCmd(checkErr, "mads", "no_name.asm", "-o:output.xex")
			// disassemble optimized picture
			disAsmRasta("output.xex", "output")
		}
		fmt.Println("See Generator folder for output and kernel files")
		time.Sleep(2 * time.Second)
	case "ext":
		fmt.Println("Extracting...")
		if !disAsmRslide(os.Args[2], os.Args[3], "output") {
			fmt.Println("Extract failed")
			os.Exit(2)
		}
		fmt.Println("Extracted OK")
		time.Sleep(2 * time.Second)
	case "opt":
		optKernel(os.Args[3], "output.png.opt", os.Args[2], true)
	}
}

func check(e error) {
	if e != nil {
		panic(e)
	}
}

func winCmd(errIgnore bool, name string, arg ...string) {
	var out []byte
	cmd := exec.Command(name, arg...)
	out, err = cmd.Output()
	if err != nil && !errIgnore {
		fmt.Println(string(out))
		os.Exit(2)
	}
}

func buildSlides(fout string) {

	// assemble all slides into xex and make atr disks

	var flist []string
	var f, picn string
	var n int

	// find existing slide pictures - which are prefixed with a 5 digit number
	// to give their order in the show. space is the separator after the number
	flist, err = filepath.Glob(filepath.Join("..", "+[0-9][0-9][0-9][0-9][0-9] *"))
	check(err)

	if len(flist) == 0 {
		fmt.Println("Add slides before building")
		os.Exit(1)
	}

	fmt.Println("Building slides...")

	for n, f = range flist {

		dat, err := ioutil.ReadFile(f)
		check(err)

		fmt.Println(filepath.Base(f))
		// copy to next pic number file
		picn = strconv.Itoa(n + 1)
		err = ioutil.WriteFile("_pic"+picn+".deflate", dat, 0644)
		// make next info number file
		writeInfo(filepath.Base(f), "_info"+picn+".asm")
	}
	// assemble slides
	fmt.Println("Assembling...")
	f = filepath.Join("dir2atr", "slides.xex")
	os.Remove(f)
	winCmd(checkErr, "mads", "slides.asm", "-o:"+f, "-d:PIC_QTY="+picn)

	// build disk

	fmt.Println("Building disk types...")

	f = strings.TrimSuffix(fout, filepath.Ext(fout))
	winCmd(checkErr, "dir2atr", "-d", "-b", "PicoBoot406", f+".atr", "dir2atr")

	// attempt to build single density disk
	winCmd(ignoreErr, "dir2atr", "-S", "-b", "PicoBoot406", f+"(SD).atr", "dir2atr")
	if err != nil {
		os.Remove(f + "(SD).atr")
	}
	// attempt to build enhanced density disk
	winCmd(ignoreErr, "dir2atr", "-E", "-b", "PicoBoot406", f+"(ED).atr", "dir2atr")
	if err != nil {
		os.Remove(f + "(ED).atr")
	}
	// attempt to build double density disk
	winCmd(ignoreErr, "dir2atr", "-D", "-b", "PicoBoot406", f+"(DD).atr", "dir2atr")
	if err != nil {
		os.Remove(f + "(DD).atr")
	}
	// attempt to build quad density disk
	winCmd(ignoreErr, "dir2atr", "-Q", "-b", "PicoBoot406", f+"(QD).atr", "dir2atr")
	if err != nil {
		os.Remove(f + "(QD).atr")
	}

	// make a non looped XEX version
	os.Rename(filepath.Join("dir2atr", "slides.xex"), f+"(no loop).xex")

	fmt.Println("Built", filepath.Base(fout))
}

func writeInfo(slide string, fout string) {

	// remove number prefix
	name := slide[7:]
	n := strings.LastIndex(name, "#")
	t := 0
	if n >= 0 {
		name = strings.TrimRight(name, "sS")
		t, _ = strconv.Atoi(name[n+1:])
		name = strings.TrimSpace(name[:n])
	}

	if t == 0 || t > 127 {
		fmt.Println("Build failed. Seconds timer can be from #1s to #127s. Rename to fix")
		os.Exit(2)
	}

	if len(name) > 40 {
		name = name[:40]
		fmt.Println("Build failed. Name over 40 characters won't fit Atari screen line. Rename to fix")
		os.Exit(2)
	}

	// try to clean up International Unicode into normalised ASCII which the
	// Atari may have some chance of displaying with the standard character set
	name = removeAccents(name)

	fo, err := os.Create(fout)
	check(err)

	fo.WriteString("\n\torg pic_info\n")
	fo.WriteString(";pic_timer\n")
	fo.WriteString("\t.byte " + strconv.Itoa(t) + "\n")
	fo.WriteString(";pic_name\n")
	fo.WriteString("\t.sb '" + name + "'\n")

	fo.Close()
}

func removeAccents(s string) string {
	// https://twinnation.org/articles/33/remove-accents-from-characters-in-go
	t := transform.Chain(norm.NFD, runes.Remove(runes.In(unicode.Mn)), norm.NFC)
	output, _, err := transform.String(t, s)
	check(err)
	return output
}

func addSlide(fin, cols string, zopts []string) bool {

	// disassemble a recognized RastaConverter XEX, process and add to slide files

	var f, name, ext string
	var flist []string
	var n, t int
	var slide bool

	ext = filepath.Ext(fin)
	name = strings.TrimSuffix(filepath.Base(fin), ext) // file name without path and extension

	if name[0] == byte('+') && ext == "" && len(name) > 7 {
		// a slide file - no conversion needed
		// remove number prefix
		name = name[7:]
		// remove timer suffix
		n = strings.LastIndex(name, "#")
		if n > 0 {
			name = strings.TrimSpace(name[:n])
		}
		slide = true
	}

	fmt.Println("Adding", name)

	// get default timer from parent folder file starting "#" ending "s (default timer)".
	flist, err = filepath.Glob(filepath.Join("..", "#*s (default timer)"))

	// I really like having the timer value in the file name, but someone is bound to
	// end up with multiple timer files somehow. Clean up will be done in that case.
	if len(flist) == 1 {
		f = filepath.Base(flist[0])
		f = strings.TrimPrefix(f, "#")
		f = strings.TrimSuffix(f, "s (default timer)")
		t, err = strconv.Atoi(f)
	}

	if t == 0 || t > 127 {
		fmt.Println("Error getting default timer. Range 1s to 127s. Recreating file '#10s (default timer)'")
		// clean up maybe multiple default timer files
		for _, f = range flist {
			os.Remove(f)
		}
		fo, _ := os.Create(filepath.Join("..", "#10s (default timer)"))
		fo.Close()
		t = 10
		time.Sleep(4 * time.Second)
	}

	// renumber the existing slide picture files and return the next slide number
	n = renumSlides()

	// make slide. prefix next number and add timer to passed file name
	f = filepath.Join("..", "+"+strconv.Itoa(n)+" "+name+" #"+strconv.Itoa(t)+"s")

	// no need to convert a slide file, just copy it
	if slide {
		dat, err := ioutil.ReadFile(fin)
		check(err)
		err = ioutil.WriteFile(f, dat, 0644)

		fmt.Println("Added OK")
		return true // success
	}

	fmt.Println("Disassembling...")
	if !disAsmRasta(fin, "output") {
		fmt.Println("Add failed. Unrecognised RastaConverter XEX")
		return false // fail
	}

	fmt.Println("Optimizing kernel...")
	optKernel("output.png.rp", "output.png.opt", cols, true)

	fmt.Println("Assembling...")
	winCmd(checkErr, "mads", "pic.asm", "-o:output.dat")

	fmt.Println("Compressing...")
	zopts = append(zopts, "output.dat")
	// zopfli has no option to overwrite an existing file
	os.Remove("output.dat.deflate")
	winCmd(checkErr, "zopfli", zopts...)
	os.Rename("output.dat.deflate", f)

	fmt.Println("Added OK")

	return true // success
}

func renumSlides() int {

	// renumber all slides and return the next slide number

	const start int = 10000 // starting number
	const step int = 10     // number slides in steps of 10 for easy manual reordering
	var flist, rlist []string
	var f, r string
	var n, i int
	var renamed bool

	// find existing slide pictures - which are prefixed with a 5 digit number
	// to give their order in the show. space is the separator after the number
	flist, err = filepath.Glob(filepath.Join("..", "+[0-9][0-9][0-9][0-9][0-9] *"))
	check(err)

	rlist = make([]string, len(flist))

	// make a list of what renumbered file names will be
	n = start
	for i, f = range flist {
		r = strconv.Itoa(n)
		// ignore files already correctly numbered
		if r != f[4:9] {

			rlist[i] = filepath.Join("..", "+"+r+f[9:])
		}
		n += step
	}
	for !renamed {
		// reset list renamed flag
		renamed = true
		for i, r = range rlist {
			// ignore files done or already correctly named
			if r == "" {
				continue
			}
			if inList(r, flist) {
				// can't rename yet, a slide of the exact same name still exists
				// flag another pass to try again until that file can itself be renamed
				renamed = false
				continue
			}
			// rename and mark as done
			os.Rename(flist[i], r)
			r = ""
			flist[i], rlist[i] = r, r
		}
	}
	// return next slide number
	return n
}

func inList(text string, list []string) bool {
	for _, s := range list {
		if s == text {
			return true
		}
	}
	return false
}

func disAsmRasta(fin, fout string) bool {

	// disassemble XEX into usual RastaConverter output files
	// bitmap .MIC file, player missile .pmg file,
	// kernel .rp code file, kernel initialize rp.ini code file

	var ok bool
	var addr, height, padln int
	var seg, bin, bmp []byte
	var wait57 []byte = []byte{
		0xea, 0xea, 0xea, 0xea,
		0xc5, 0x0a,
		0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea,
		0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea,
		0xea, 0xea, 0xea,
	}

	xex, err := ioutil.ReadFile(fin)
	check(err)

	// get 1st segment
	pos = 0
	addr, seg = nextSeg(xex)

	if addr != 0x2000 || len(seg) != 6 {
		return false
	}

	// assume RastaConverter ini code

	// discard ini run address
	nextSeg(xex)

	// bitmap
	addr, bmp = nextSeg(xex)
	height = len(bmp) / 40

	writeMIC(bmp, height, fout)

	// discard DLIST
	nextSeg(xex)

	// segment is normally player data and program
	// or player line padding with Drpeter centre mod
	addr, seg = nextSeg(xex)

	// player line padding segment must be shorter than the screen height
	// player data and program segment is always longer than that
	if len(seg) < 240 {
		// Drpeter centre mod
		fmt.Println("Player padding segment")
		padln = len(seg)
		// player data
		addr, seg = nextSeg(xex)
		// include line padding in the player data
		seg = append(bytes.Repeat([]byte{0}, padln), seg...)
		// player data only 1024 bytes
		writePMG(seg[:1024], fout)
		// program segment
		addr, seg = nextSeg(xex)
	} else {
		// player data 1024 bytes
		writePMG(seg[:1024], fout)
		// discard player data leaving just program in segment
		seg = seg[1024:]
		addr = addr + 1024
	}

	// disassemble ini

	// find sta DLISTHI before ini
	start := bytes.Index(seg, []byte{0x8d, 0x03, 0xd4}) + 3

	// find the WSYNCs ($D40A) at end of .ini (sta $D40A \ sta $D40A)
	end := bytes.Index(seg, []byte{0x8d, 0x0a, 0xd4, 0x8d, 0x0a, 0xd4}) + 6

	if start >= end {
		return false
	}

	ini, _, ok := disAsmKernel(seg[start:end], false)

	writeRPini(addr+start, ini, height, fout)

	if !ok {
		return false
	}

	// disassemble raster program kernel

	// 19 bytes of 100 wait cycles after kernel ini and before rp kernel
	//
	// --- wait 18 cycles
	// jsr _rts
	// inc byt3
	//
	// --- set global offset (23 cycles)
	// jsr _rts
	// cmp byt3\ pha:pla
	//
	// --- empty line
	// jsr wait54cycle
	// inc byt2
	//
	// if drpeter centre mod there are pad lines of wait cycles immediately
	// following, before what goes in the raster program kernel (.rp) file
	//
	// jsr wait57cycle
	// jsr wait57cylce
	// ...

	// find (nop \ nop \ nop \ nop \ cmp zp) at start of raster program
	start = end + bytes.Index(seg[end:], []byte{0xea, 0xea, 0xea, 0xea, 0xc5})

	// do not do kernel line padding unless there are extra wait cycles
	if start-end <= 19 {
		padln = 0
	}

	// find jump to end of kernel address before loop start and wait for VCOUNT ($D40B)
	// jmp raster_program_end \ LOOP \ lda VCOUNT \ cmp #$02
	pos = bytes.Index(seg, []byte{0x0b, 0xd4}) - 3

	// convert jump address to slice index
	end = int(seg[pos]) + int(seg[pos+1])<<8 - addr

	if start >= end {
		return false
	}

	bin = seg[start:end]

	// insert pad line wait cycles to replace jsr wait57cycle lines
	if padln > 0 {
		bin = append(bytes.Repeat(wait57, padln), bin...)
	}

	kernel, _, ok := disAsmKernel(bin, true)

	writeRP(addr+start, kernel, fout)

	if !ok {
		return false
	}

	return true // success, probably
}

func nextSeg(xex []byte) (int, []byte) {

	var addr, start, end int

	// Atari binary load format
	// ------ header ----------
	// 0xffff	word
	// ------ segment ---------
	// addr start	word
	// addr end	word
	// data	...	end-start+1
	// ------ segment ---------
	// addr start	word
	// addr end	word
	// data ...	end-start+1

	if pos == 0 {
		start = int(xex[pos]) + int(xex[pos+1])<<8
		pos += 2
		if start != 0xffff {
			fmt.Println("Not Atari 8-bit program")
			return -1, []byte{}
		}
	}
	if pos >= len(xex) {
		// end of file
		return -1, []byte{}
	}
	addr = int(xex[pos]) + int(xex[pos+1])<<8
	start = pos + 4
	end = start + int(xex[pos+2]) + int(xex[pos+3])<<8 - addr + 1
	pos = end

	return addr, xex[start:end]
}

func writeMIC(bmp []byte, height int, fout string) {

	// convert of ilmenit_orca
	dat, err := os.Create(fout + ".png.dat")
	defer dat.Close()
	check(err)
	dat.Write(bmp)

	mic, err := os.Create(fout + ".png.mic")
	defer mic.Close()
	check(err)
	// when writing extracted RastaSlide padded bitmap, only do up to actual height
	if height >= 204 {
		// MIC format does not include 16 byte padding for last LMS boundary
		mic.Write(append(bmp[:8160], bmp[8160+16:height*40+16]...))
	} else {
		mic.Write(bmp[:height*40])
	}
}

func writePMG(player []byte, fout string) {
	pg, err := os.Create(fout + ".png.pg")
	defer pg.Close()
	check(err)
	pg.Write(player[:1024])
	pmg, err := os.Create(fout + ".png.pmg")
	defer pmg.Close()
	check(err)
	pmg.WriteString("\t.ds $100\n")
	pmg.WriteString("\tins \"" + filepath.Base(fout+".png.pg") + "\"\n")
}

func writeRPini(addr int, ini string, height int, fout string) {
	// write kernel ini file
	rpini, err := os.Create(fout + ".png.rp.ini")
	defer rpini.Close()
	check(err)
	if addr >= 0 {
		rpini.WriteString("; org $" + fmt.Sprintf("%04X", addr) + "\n")
	}
	rpini.WriteString(ini)
	// write picture height label
	rpini.WriteString("\nPIC_HEIGHT = " + strconv.Itoa(height) + "\n")
}

func writeRP(addr int, kernel string, fout string) {
	// write kernel file
	rp, err := os.Create(fout + ".png.rp")
	defer rp.Close()
	check(err)
	if addr >= 0 {
		rp.WriteString("; org $" + fmt.Sprintf("%04X", addr) + "\n")
	}
	rp.WriteString(kernel)
}

func disAsmKernel(bin []byte, kernel bool) (string, int, bool) {

	// disassemble RastaConverter program ini or kernel

	const cycln = 57            // cycles per line in kernel
	var cyctot int = cycln - 11 // line 0 label after 11 cycles in rp file
	var op, pc, height int
	var val, hpos uint64
	var asm, addr, code, pix, warn, ind string
	var opmap = []byte{0xa9, 0xa2, 0xa0, 0x8d, 0x8e, 0x8c, 0xea, 0xc5}
	var inst = []string{
		"lda", "ldx", "ldy",
		"sta", "stx", "sty",
		"nop\t", "cmp byt2",
	}
	var mode = []int{1, 1, 1, 2, 2, 2, 0, 0}
	var size = []int{2, 2, 2, 3, 3, 3, 1, 2}
	var cyc = []int{2, 2, 2, 4, 4, 4, 2, 3}
	var reg = []int{1, 2, 3, 1, 2, 3, 0, 0}
	var colclk = []uint64{
		217, 219, 221, 223, 225, 15, 17, 19, 21, 23,
		25, 27, 29, 31, 33, 35, 37, 41, 45, 53,
		61, 69, 77, 85, 93, 101, 109, 117, 121, 125,
		129, 133, 137, 141, 145, 149, 153, 157, 161, 165,
		169, 173, 177, 181, 185, 189, 193, 197, 199, 201,
		203, 205, 207, 209, 211, 213, 215,
	}
	var labmap = map[string]string{
		string([]byte{0x16, 0xd0}): "COLPF0",
		string([]byte{0x17, 0xd0}): "COLPF1",
		string([]byte{0x18, 0xd0}): "COLPF2",
		string([]byte{0x1a, 0xd0}): "COLBAK",
		string([]byte{0x12, 0xd0}): "COLPM0",
		string([]byte{0x13, 0xd0}): "COLPM1",
		string([]byte{0x14, 0xd0}): "COLPM2",
		string([]byte{0x15, 0xd0}): "COLPM3",
		string([]byte{0x00, 0xd0}): "HPOSP0",
		string([]byte{0x01, 0xd0}): "HPOSP1",
		string([]byte{0x02, 0xd0}): "HPOSP2",
		string([]byte{0x03, 0xd0}): "HPOSP3",
		string([]byte{0x1e, 0xd0}): "HITCLR",
		string([]byte{0x0a, 0xd4}): "WSYNC",
	}
	for pc < len(bin) {
		op = bytes.Index(opmap, bin[pc:pc+1])
		if op < 0 {
			code += fmt.Sprintf("opcode $%02X?", bin[pc:pc+1])
			fmt.Println("Unexpected opcode in RastaConverter kernel")
			return code, height, false
		}
		addr = ""
		ind = ""
		if kernel {
			// label line numbers in kernel code
			if cyctot >= cycln {
				code += fmt.Sprintf("line%d\t\t\t; VPOS=%d\n", height, height+8)
				cyctot -= cycln
				height++
			}
			cyctot += cyc[op]
			hpos = colclk[cyctot%cycln]
		}
		asm = "\t" + inst[op]

		switch mode[op] {
		case 1:
			// 8-bit immediate value after instruction
			asm += fmt.Sprintf(" #$%02X\t;\t", bin[pc+1])
			regval[reg[op]] = uint64(bin[pc+1])
		case 2:
			// 16-bit address after instruction
			addr = labmap[string(bin[pc+1:pc+3])]
			if addr != "" {
				asm += " " + addr
			} else {
				asm += fmt.Sprintf(" $%02X%02X", bin[pc+2], bin[pc+1])
			}
			val = regval[reg[op]]
			if strings.HasPrefix(addr, "HPOS") && val-hpos > 0 && val-hpos <= 5 {
				ind = "<--"
				warn += strconv.Itoa(height-1) + " "
			}
			asm += fmt.Sprintf("\t; $%02X(%d)", val, val)
		default:
			asm += "\t;\t"
		}
		if kernel {
			pix = ""
			if hpos >= 0+48 && hpos < 160+48 {
				pix = "PIX=" + strconv.Itoa(int(hpos)-48)
			}
			asm += fmt.Sprintf("\t@%02d HPOS=%d%v\t%v", cyctot, hpos, ind, pix)
		}
		code += asm + "\n"
		pc += size[op]
	}
	if warn != "" {
		warn = "; <-- possible artefacts from 5 colour clock HPOS delay at:\n; line" + warn + "\n"
		code = warn + code
	}
	return code, height, true
}

func disAsmRslide(opt, fin, fout string) bool {

	// extract slides and recreate RastaConverter pictures from slide show disk

	const step int = 10 // number slides in steps of 10 for easy manual reordering
	var sect, sectsize, start, end, addr, n, t int
	var f, name, ext string
	var atr, xex, seg []byte

	ext = strings.ToLower(filepath.Ext(fin))

	if ext == ".atr" {
		//
		// extract XEX file from ATR DOS disk
		//

		// 16 byte ATR header (16-bit words are little endian)
		// ---------------------------------------------------
		// +0	 	word $0296 magic number (sum of "NICKATARI")
		// +2		word disk size in 16 byte paragraphs
		// +4		word sector size
		// +6		byte disk size high paragraphs
		// +7-15	various

		atr, err = ioutil.ReadFile(fin)

		sectsize = int(atr[4]) + int(atr[5])<<8

		// remove the header and the 3 boot sectors of 128 bytes (128 regardless of sector size)
		atr = atr[3*128+16:]

		// check we have a DOS disk with slides starting at sector 4 (sectors numbered from 1)
		// VTOC at sector 360, then directory starting at sector 361
		// file entries: [flag] [size in sectors low][high] [start sector high][low] [N][A][M][E]...
		start = (361-4)*sectsize + 3
		end = start + 8

		if end >= len(atr) {
			return false // definitely not DOS disk
		}

		// check file number zero starts at sector 4 and has a name beginning SLIDES...
		if !bytes.Equal(atr[start:end], []byte{04, 00, 0x53, 0x4C, 0x49, 0x44, 0x45, 0x53}) {
			return false // unrecognised disk
		}

		sect = 4

		// last 3 bytes of each DOS sector [file no | next sector high][low] [bytes used]
		for sect > 0 {
			start = (sect - 4) * sectsize
			end = start + int(atr[start+sectsize-1])
			xex = append(xex, atr[start:end]...)
			sect = int(atr[start+sectsize-3])<<8 + int(atr[start+sectsize-2]) // NB big endian
		}

		// output the non looped XEX as a file in case anyone is interested
		slides, err := os.Create(filepath.Join("..", "RastaSlides(no loop).xex"))
		defer slides.Close()
		check(err)
		slides.Write(xex)

	} else {
		xex, err = ioutil.ReadFile(fin)
	}

	//
	// disassemble extracted XEX file
	//

	// get 1st segment
	addr, seg = nextSeg(xex)

	if addr == 0x2000 && (len(seg) == 82 || len(seg) == 78) {
		// assume RastaSlide executable make_dlist code

		// discard make_dlist ini
		nextSeg(xex)
		// discard viewer code
		nextSeg(xex)

		if opt == "sld" {
			// renumber the existing slide picture files and return the next slide number
			n = renumSlides()
		}

		// slides
		for {
			// discard info ini address but check for end of slides.
			// run address to warm start is set after the last slide
			addr, seg = nextSeg(xex)
			if addr == 0x2e0 || addr < 0 {
				break // end of pictures in xex
			}

			// timer and name
			addr, seg = nextSeg(xex)

			t = int(seg[0])

			// discard timer
			seg = seg[1:]

			// convert Atari screen codes to ASCII
			for i, _ := range seg {
				if seg[i] < 64 {
					seg[i] = seg[i] + 32
				}
			}
			name = string(seg)
			fmt.Println(name)

			// compressed picture data
			addr, seg = nextSeg(xex)

			if opt == "sld" {

				// make slide. prefix next number and add timer to name
				f = filepath.Join("..", "+"+strconv.Itoa(n)+" "+name+" #"+strconv.Itoa(t)+"s")

				def, err := os.Create(f)
				defer def.Close()
				check(err)
				def.Write(seg)

				n += step

			} else if opt == "pic" {

				def, err := os.Create(fout + ".dat.deflate")
				defer def.Close()
				check(err)
				def.Write(seg)

				// decompress picture data
				decompress(fout+".dat.deflate", fout+".dat")

				// get decompressed data
				dat, err := ioutil.ReadFile(fout + ".dat")
				check(err)

				// player data 1024 bytes
				writePMG(dat[:1024], fout)

				// bitmap data (padded to maximum possible height, not actual picture height)

				dat = dat[1024+16:]    // remove player data and 16 byte screen alignment
				bmp := dat[:240*40+16] // 240*40=9600 + 16 for 2nd 4K boundary padding

				// disassemble ini

				dat = dat[240*40+16:] // remove bitmap data
				start = 0

				// find the WSYNCs ($D40A) at end of .ini (sta $D40A \ sta $D40A)
				end = bytes.Index(dat, []byte{0x8d, 0x0a, 0xd4, 0x8d, 0x0a, 0xd4}) + 6

				ini, _, _ := disAsmKernel(dat[start:end], false)

				// disassemble raster program

				// skip 100 cycle wait (50 bytes) (nop \ nop \ nop ...)
				start = end + 50

				// don't include the rts at the end
				end = len(dat) - 1

				kernel, height, _ := disAsmKernel(dat[start:end], true)

				writeRPini(-1, ini, height, fout)
				writeRP(-1, kernel, fout)

				writeMIC(bmp, height, fout)

				// build standard RastaConverter XEX of slide name
				winCmd(checkErr, "mads", "no_name.asm", "-o:"+filepath.Join("..", name)+".xex")
			}

			// discard inflate addresses
			nextSeg(xex)
			// discard inflate ini
			nextSeg(xex)
			// discard viewer ini
			nextSeg(xex)
		}
		return true // success
	}
	return false // failed
}

func decompress(fin, fout string) {
	i, _ := os.Open(fin)
	defer i.Close()
	f := flate.NewReader(i)
	defer f.Close()
	o, _ := os.Create(fout)
	defer o.Close()
	io.Copy(o, f)
}

func optKernel(fin, fout, cols string, defrag bool) {

	// optimize raster kernel source code file by de-duplicating redundant register loads and stores.
	// if defrag is true, also optimize for size after assembly and LZ compression (or RLE)
	// assumes input file has associated ini of same name with .ini appended.
	// output file created with an associated ini with optimized initial register values.
	// only saves 6 bytes doing the optimized ini register values, so not sure the extra complexity
	// in this function is really worth it, but it's done now. the other optimizations save much more
	// but no data transforms are done here (that require reversing on the Atari after decompression)

	const nop = "\tnop\n"
	var regchk int
	var asm, bubble string
	var swapln int
	var cmpln int
	var ldln []int = make([]int, 4)
	var undoln []int = make([]int, 4)
	var undo []string = make([]string, 4)

	// RastaConverter actually outputs 256 colour values, but only 128 colours are visible
	// unless using a Sophia GTIA replacement in non-standard enhanced mode. Better optimization
	// is possible in the kernel if the invisible brightness values are converted to the visible.
	// Using even lower brightness precision values will be visible but could be interesting,
	// but lower than 32 we don't even have on/off brightness for a colour without more code
	switch cols {
	case "256":
		colmask = 0b11111111
	case "128":
		colmask = 0b11111110
	case "64":
		colmask = 0b11111100
	case "32":
		colmask = 0b11111000
	default:
		colmask = 0b11111110
	}

	// get kernel .ini file
	dat, err := ioutil.ReadFile(fin + ".ini")
	check(err)

	ini := strings.SplitAfter(string(dat), "\n")

	// get kernel source file
	dat, err = ioutil.ReadFile(fin)
	check(err)

	kernel := strings.SplitAfter(string(dat), "\n")

	// merge ini and kernel code so initial register values can be optimized simply.
	// will get split out again at the end
	code := append(ini, kernel...)

	//
	// de-dupe code
	//

	// remove redundant load and stores. size optimize by replacing with NOPs,
	// including filler stores (HITCLR) and unrecognised stores not in addrmap or strbmap

	for pass := 1; pass <= 3; pass++ {
		// reset registers and hardware addresses to take new values that won't get optimized away
		// by forcing inequality with any 8-bit value
		for n, _ := range regval {
			regval[n] = 0xffffffff
		}
		for n, _ := range addrval {
			addrval[n] = 0xffffffff
		}
		for n := 0; n < len(code); n++ {
			parseLine(code[n])
			if ld {
				if pass == 2 {
					// save original load line
					undo[reg] = code[n]
					undoln[reg] = n
					// colour mask load line value
					asm = code[n][:strings.Index(code[n], "#")+1] + "$"
					asm += fmt.Sprintf("%02X\n", regval[reg]&colmask)
					code[n] = asm
				}
				if regval[reg] == regold[reg] && n > len(ini) {
					// optimize redundant load but ignore initial loads to addresses
					code[n] = nop

				} else if defrag || n > len(ini) {
					// only can deal with redundancy in initial loads to registers when defragging
					if ldln[reg] > 0 {
						// optimize previous load line of register
						// 2nd pass will remove any redundant loads from end of 1st pass
						code[ldln[reg]] = nop
					}
					// assume load line is redundant. only last one confirmed by a store counts
					ldln[reg] = n
				}
			} else if st {
				if pass == 2 && hpos {
					// undo colour masked load if used for a player position
					// NB redundant stores can spoil optimization so must be done after pass 1
					code[undoln[reg]] = undo[reg]
				}
				if addrval[addr] == addrold[addr] || addr == 0 {
					// NB colour mask already applied to colour addresses in parseLine
					// optimize redundant store or filler store
					code[n] = nop
					// insert an extra copy of the nop line to match store cycles
					code = append(code, "")
					copy(code[n+1:], code[n:])
				} else {
					// load line of register has a store. stop line from being optimized
					ldln[reg] = 0
				}
			}
		}
	}

	if defrag {

		//
		// Defrag NOPs for better run and LZ compression
		//

		// a) move loads up to their register's previous store (including into unused initial values)

		for {
			bubble = ""
			for n := 1; n < len(code); n++ {
				parseLine(code[n])
				if ld {
					swapln = n
					regchk = reg
					for m := n - 1; m > 0; m-- {
						if strings.HasPrefix(code[m], "\tnop") {
							swapln = m
						} else {
							parseLine(code[m])
							if reg == regchk {
								break // done moving up - next line up is same register
							}
						}
					}
					if swapln != n {
						bubble = code[n]
						code[n] = code[swapln]
						code[swapln] = bubble
					}
				}
			}
			if bubble == "" {
				break
			}
		}

		// b) replace end of line wait CMPs with 3 NOPs if there are 2 of them in a block of NOPs

		// leave 1st CMP in the kernel for more compatible RastaConverter if extracted (and easier kernel detect)
		for n := len(ini) + 5; n < len(code); n++ {
			parseLine(code[n])
			if ld || st {
				// flag no longer in wait block
				cmpln = 0
			} else if strings.HasPrefix(code[n], "\tcmp") {
				if cmpln == 0 {
					// 1st cmp line
					cmpln = n
				} else {
					// 2nd cmp line in wait block
					// replace the 2 cmp lines with nops
					code[n] = nop
					code[cmpln] = nop
					// insert an extra copy of the nop line to match cmp zp cycles
					code = append(code, "")
					copy(code[n+1:], code[n:])
					cmpln = 0
				}
			}
		}

		// c) move end of line wait CMPs to the start of the wait filler NOPs

		for n := 1; n < len(code); n++ {
			parseLine(code[n])
			if strings.HasPrefix(code[n], "\tcmp") {
				swapln = n
				// do not swap 1st CMP into ini file (for more compatible RastaConverter if extracted)
				for m := n - 1; m >= len(ini)+5; m-- {
					// find first nop in wait filler block skipping comments and labels
					if strings.HasPrefix(code[m], "\tnop") || !parseLine(code[m]) {
						swapln = m
					} else {
						break // done moving up - next line up is code that isn't NOP
					}
				}
				if swapln != n {
					bubble = code[n]
					code[n] = code[swapln]
					code[swapln] = bubble
				}
			}
		}
	}

	//
	// output optimized ini and kernel files
	//

	rpini, err := os.Create(fout + ".ini")
	defer rpini.Close()
	check(err)

	for n := 0; n < len(ini); n++ {
		rpini.WriteString(code[n])
	}

	rp, err := os.Create(fout)
	defer rp.Close()
	check(err)

	for n := len(ini); n < len(code); n++ {
		rp.WriteString(code[n])
	}
}

func parseLine(line string) bool {

	fields := strings.Fields(line)
	ld, st, wt = false, false, false
	fld, reg, addr = 0, 0, 0

	if len(fields) == 0 {
		// blank line
		return false
	}
	text := nextField(fields)
	if text[0] == byte(';') {
		// comment
		return false
	}
	// instruction
	parseField1(text)
	// check for label (or :repeat)
	if !(ld || st || wt) {
		text = nextField(fields)
		parseField1(text)
	}
	if !(ld || st || wt) {
		// no instruction following label (or :repeat)
		return false
	}
	// operand, (will include the ";" from "byt2;" as no white space)
	text = nextField(fields)
	parseField2(text)

	return true
}

func nextField(fields []string) string {
	if len(fields) > fld {
		fld++
		return fields[fld-1]
	} else {
		return " "
	}
}

func parseField1(text string) {
	var regmap = map[string]int{
		"lda": 1,
		"ldx": 2,
		"ldy": 3,
		"sta": 1,
		"stx": 2,
		"sty": 3,
	}
	var ldmap = map[string]bool{
		"lda": true,
		"ldx": true,
		"ldy": true,
	}
	var stmap = map[string]bool{
		"sta": true,
		"stx": true,
		"sty": true,
	}
	var wtmap = map[string]bool{
		"nop": true,
		"cmp": true,
	}

	reg = regmap[text]
	ld = ldmap[text]
	st = stmap[text]
	wt = wtmap[text]
}

func parseField2(text string) {
	var addrmap = map[string]int{
		"COLOR0": 1,
		"COLPF0": 1,
		"COLOR1": 2,
		"COLPF1": 2,
		"COLOR2": 3,
		"COLPF2": 3,
		"COLBAK": 4,
		"COLBK":  4,
		"COLPM0": 5,
		"COLPM1": 6,
		"COLPM2": 7,
		"COLPM3": 8,
		"HPOSP0": 9,
		"HPOSP1": 10,
		"HPOSP2": 11,
		"HPOSP3": 12,
	}
	var strbmap = map[string]bool{
		"WSYNC": true,
		"wsync": true,
	}

	if text[0] == byte('#') {
		// load immediate
		regold[reg] = regval[reg]
		if text[1] == byte('$') {
			regval[reg], err = strconv.ParseUint(text[2:], 16, 8)
		} else {
			// RastaConverter outputs in Hex, but if there are any hand modifications in decimal
			regval[reg], err = strconv.ParseUint(text[1:], 10, 8)
		}
		check(err)
	} else if st {
		// store absolute
		addr = addrmap[text]
		// only trace valid hardware addresses. anything else except for
		// store to strobes (WSYNC) will get optimized away.
		if addr > 8 {
			addrold[addr] = addrval[addr]
			addrval[addr] = regval[reg]
			// mustn't use colour mask on the load register as a player hpos has been written
			hpos = true
		} else if addr > 0 {
			addrold[addr] = addrval[addr]
			// RastaConverter actually outputs 256 colour values, but only 128 colours can display
			// on standard Atari hardware. The colour mask is used to optimize that when wanted.
			// This ensures that colour registers don't get written again if only bit 0 changes.
			addrval[addr] = regval[reg] & colmask
			hpos = false
		} else if strbmap[text] {
			// leave addrold and addrval different so strobe store doesn't get optimized.
			// any register loads stored just to the strobe are to be optimized though.
			// this is all a bit of overkill for WSYNCs at the end of the ini file that
			// could have just been made a hacked exception and treated like a comment
			st = false
			reg = 0
		}
	}
}
